/*
		complex functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Complex.h"

/* add two complex numbers */
Complex CAdd(Complex a, Complex b)
{
	Complex s;
	
	s.real = a.real + b.real;
	s.imag = a.imag + b.imag;
	
	return s;
}

/* subtract two complex numbers */
Complex CSub(Complex a, Complex b)
{
	Complex d;
	
	d.real = a.real - b.real;
	d.imag = a.imag - b.imag;
	
	return d;
}

/* multiply two complex numbers */
Complex CMul(Complex a, Complex b)
{
	Complex p;

	p.real = a.real * b.real - a.imag * b.imag;
	p.imag = a.imag * b.real + a.real * b.imag;	
	
	return p;
}

/* divide two complex numbers */
Complex CDiv(Complex a, Complex b)
{
	double d;
	Complex q;
	
	d = b.real * b.real + b.imag * b.imag;
	
	q.real = (a.real * b.real + a.imag * b.imag) / d; 
	q.imag = (a.imag * b.real - a.real * b.imag) / d; 		
	
	return q;
}

/* negate complex number */
Complex CNeg(Complex z)
{
	z.real = -z.real;
	z.imag = -z.imag;
	
	return z;
}

/* conjugate of complex number */
Complex CConj(Complex a)
{
	a.imag = -a.imag;
	
	return a;
}

/* magnitude of complex number */
double CAbs(Complex a)
{
	return sqrt (a.real * a.real + a.imag * a.imag);
}

/* phase of complex number */
double CAngle(Complex a)
{
	return atan2 (a.imag, a.real);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* multiply real with complex */
Complex CMulReal(double r, Complex c)  
{
	c.real *= r;
	c.imag *= r;
	return c;
}

/* divide complex by real*/
Complex CDivReal(double r, Complex c)  
{
	c.real /= r;
	c.imag /= r;
	return c;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* convert double to complex */
Complex Double2Comp(double x)
{
	Complex z;
	
	z.real = x;
	z.imag = 0.;
	
	return z;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* complex arc cosine */
Complex CACos(Complex z)
{
	Complex r;
	
	cacos (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic arc cosine */
Complex CACosH(Complex z)
{
	Complex r;
	
	cacosh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex arc sine */
Complex CASin(Complex z)
{
	Complex r;
	
	casin (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic arc sine */
Complex CASinH(Complex z)
{
	Complex r;
	
	casinh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex arc tangent */
Complex CATan(Complex z)
{
	Complex r;
	
	catan (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic arc tangent */
Complex CATanH(Complex z)
{
	Complex r;
	
	catanh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex cosine */
Complex CCos(Complex z)
{
	Complex r;
	
	ccos (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic cosine */
Complex CCosH(Complex z)
{
	Complex r;
	
	ccosh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex exponential */
Complex CExp(Complex z)
{
	Complex r;
	
	cexp (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex natural logarithm (base e) */
Complex CLn(Complex z)
{
	Complex r;
	
	cln (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex decadic logarithm (base 10) */
Complex CLog10(Complex z)
{
	Complex r;
	
	clog10 (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}
 
/* complex power function x^y
https://de.wikipedia.org/wiki/Komplexe_Zahl#Beliebige_komplexe_Exponenten */
Complex CPow(Complex x, Complex y)
{
	return CExp(CMul(y, CLn(x)));
}

/* complex sine */
Complex CSin(Complex z)
{
	Complex r;
	
	csin (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic sine */
Complex CSinH(Complex z)
{
	Complex r;
	
	csinh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex square root */
Complex CSqrt(Complex z)
{
	Complex r;
	
	csqrt (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex tangent */
Complex CTan(Complex z)
{
	Complex r;
	
	ctan (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}

/* complex hyperbolic tangent */
Complex CTanH(Complex z)
{
	Complex r;
	
	ctanh (z.real, z.imag, &r.real, &r.imag);
	
	return r;
}
