 /*
		debugging functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Complex.h"
#include "Matrix.h"
#include "ROM.h"
#include "Stack.h"
#include "Tags.h"

/* 2D matrix with dynamic size demo */
/*void Mat2DDemo(void)
{
	HANDLE arrHdl;
	double *arrPtr;
	short rows = 2, cols = 3, i, j;
	 
	arrHdl = HeapAllocThrow(rows * cols * sizeof(double));
	 
	arrPtr = HeapDeref (arrHdl);
	 
	for (i = 0; i < rows; i++) {
        for (j = 0; j < cols; j++) {
            arrPtr[IDX2D (cols, i, j)] = (i + 1) * (j + 1);
        }
	}
	 
	push_quantum (END_TAG);
	 
	for (i = rows - 1; i >= 0; i--) {
		push_quantum (END_TAG);

        for (j = cols - 1; j >= 0; j--) {
            push_Float (arrPtr[IDX2D (cols, i, j)]);
        }
        
		push_quantum (LIST_TAG);
	}
	 
	push_quantum (LIST_TAG);
	
	HeapFree (arrHdl); 
}*/

/* pop and push a matrix */
/*void MatPassThru(void)
{
    Access_AMS_Global_Variables;  

	EStackIndex argptr = top_estack;
	volatile HANDLE xHdl;
	double *xPtr;
	unsigned short rows = 0, cols = 0;

	// check number of parameters
	CheckNumArgs (1);		

	TRY		 
		// read list into array
		GetFloatMatArg (argptr, 
			&xHdl,
			&rows,
			&cols);
				
		// list is not empty
		if (xHdl != H_NULL)
		{
			xPtr = HeapDeref (xHdl);
			
			// push list in reverse order to estack
			push_float_mat (xPtr, rows, cols);
	  	}	
	FINALLY
		HeapFreeIndir ((HANDLE *)&xHdl);
	ENDFINAL
}*/

/* pop and push a complex matrix */
/*void ComplexMatPassThru(void)
{
    Access_AMS_Global_Variables;  

	EStackIndex argptr = top_estack;
	volatile HANDLE xHdl;
	Complex *xPtr;
	unsigned short rows = 0, cols = 0;

	// check number of parameters
	CheckNumArgs (1);		

	TRY		 
		// read list into array
		GetComplexFloatMatArg (argptr, 
			&xHdl,
			&rows,
			&cols);
				
		// list is not empty
		if (xHdl != H_NULL)
		{
			xPtr = HeapDeref (xHdl);
			
			// push list in reverse order to estack
			push_complex_float_mat (xPtr, rows, cols);
	  	}	
	FINALLY
		HeapFreeIndir ((HANDLE *)&xHdl);
	ENDFINAL
}*/

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* push expression to which pointer in array is pointing */
void PushArrExpr(EStackIndex arr[], 
	unsigned short i)
{
	push_expression (arr[i]);
}

/* push name and code of tag 
??? is this wrapper really necessary */
static void PushNameCode(EStackIndex tag)
{
	// output buffer
	char out[SYM_LEN + 5];     

	sprintf (out, "%s %d", (char *)Tag2Str (ESTACK (tag)), ESTACK (tag));
	push_zstr (out);
}

/* push row with description of quantum */
static void PushRow (Quantum q, 
	EStackIndex i,
	Boolean pushQuantum)
{
	// push line with two columns for primary tag
	push_quantum (END_TAG);
	
	if (pushQuantum) {
		// the primary tag has no value
		push_quantum (q); 
	} else {
        // value of expression
		push_expression (i); 			  	        
	}
	
	// push name and code of tag
	PushNameCode (i);
	push_quantum (LIST_TAG);	
}

/* push contents on stack as matrix to top of stack 
["NAME_OF_TAG CODE_OF_TAG" VALUE_OF_EXPR;...] 
theoretically we could save a lot of memory by removing the arrays of strings 
but when using get_key_ptr there are a lot of tags missing (this affects 
primary, secondary and command tags also) */
void PushStackAsMat(void)
{
    Access_AMS_Global_Variables;  

	// start with first arg
	EStackIndex tag = top_estack;
	
	// end of matrix
	push_quantum (END_TAG);
	  
	// loop over tags on stack from first to last arg
	while (tag >= bottom_estack) {
		switch (ESTACK (tag)) {  
			// handle all these cases with push_expresion so we do not need
			// to know their representation in detail and also do not need 
			// temp vars to store their contents between popping and pushing them
			// this way we also get support for arbitrary size ints and fractions
			// with a minimum of code
			case VAR_TAG:
			case SYSVAR_TAG:
			case ARB_REAL_TAG:
			case ARB_INT_TAG:
			case NONNEGATIVE_INTEGER_TAG:			
			case NEGATIVE_INTEGER_TAG:
			case POSITIVE_FRACTION_TAG:
			case NEGATIVE_FRACTION_TAG:
			case FLOAT_TAG: 
			case STR_DATA_TAG:
				// first arg is arbitrary in this case
				PushRow (UNDEFINED_TAG, tag, FALSE);	
				
				// advance to next tag/expression
				tag = next_expression_index (tag);
				break; 
			default: 
				PushRow (UNDEFINED_TAG, tag, TRUE);	

				// previously pushed tag was an extended tag
				// here we specify which extended was found
				if (ESTACK (tag) == SECONDARY_TAG) {
					// advance to extended tag/expression
					tag--;

					PushRow (UNDEFINED_TAG, tag, TRUE);	
				}

				// advance to next tag/expression
				tag--;
		}
	}
  
	// beginning of matrix
	push_quantum (LIST_TAG);
	
	// at least one hit was found 
	// (otherwise we would have an empty list on top of estack)
	if (is_matrix (top_estack)) {
		// just to improve readablity by horizontal scrolling
		push_transpose_aux (top_estack, FALSE);	
		
		// remove mat before transposition
		delete_expression (next_expression_index (top_estack));   
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* also see ROM call push_var */
void push_zstr_as_var(char *str)
{
	push_quantum (VAR_TAG);
	push_zstr (str);
	push_quantum (VAR_TAG);
}

#define FILE "tmp"

/* prompt user for string 
??? maybe VarREacall is better solution */
void InputStr(char *prompt,
	char buf[],
	unsigned short len)
{
	Access_AMS_Global_Variables;

	FILES file;
	
	TRY
		// prompt user for string and store string in file
		// ??? the expected variable is not created
		push_zstr_as_var (FILE);
		push_zstr (prompt);
		cmd_inputstr (top_estack);

		// file exists and could be opened
		// ??? maybe type is wrong
		if (FOpen (FILE, &file, FM_READ, "STR") == FS_OK) {
			// get string from file
			FRead (buf, len, &file);
			FClose (&file);
		} else {
			buf[0] = '\0';
		}
	FINALLY
		FDelete (FILE);
		delete_between (next_expression_index (top_estack), top_estack);
	ENDFINAL
}

///////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef DEBUG

/* display expression array */
void ExprArrPause(EStackIndex arr[], 
	unsigned short len)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_expression_list (arr, len);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display expression matrix */
void ExprMatPause(EStackIndex *mat, 
	unsigned short rows,
	unsigned short cols)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_expression_mat (mat, rows, cols);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display float array */
void FloatArrPause(double arr[], 
	unsigned short len)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_float_list (arr, len);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display float matrix */
void FloatMatPause(double *mat, 
	unsigned short rows,
	unsigned short cols)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_float_mat (mat, rows, cols);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display complex float array */
void CFloatArrPause(Complex arr[], 
	unsigned short len)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_complex_float_list (arr, len);
	replace_top_with_post_simplified (top_estack);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display complex float matrix */
void CFloatMatPause(Complex *mat, 
	unsigned short rows,
	unsigned short cols)
{
	Access_AMS_Global_Variables;
	
	EStackIndex old_top = top_estack;
	
	push_complex_float_mat (mat, rows, cols);
	replace_top_with_post_simplified (top_estack);
	cmdpause (top_estack);
	top_estack = old_top;
}

/* display name of tag */
void TagPause(EStackIndex i)
{
	Access_AMS_Global_Variables;
	
	push_zstr (Tag2Str (ESTACK (i)));
	cmdpause (top_estack);
	delete_expression (top_estack);
}

/* display SYM_ENTRY */
void SYM_ENTRYPause(SYM_ENTRY se)
{
	Access_AMS_Global_Variables;
	EStackIndex old_top = top_estack; 
	MULTI_EXPR *ptr;
	char buf[30];
	    
    push_quantum (END_TAG);
    	
	ptr = HeapDeref (se.hVal);
	
	// content of MULTI_EXPR
	// ??? gives internal error for basic funcs/progs
	//push_expression (HToESI (se.hVal)); 
    
	// pointer to MULTI_EXPR
	// ??? pushes zero
	//sprintf (buf, "Ptr %p", (void *)&ptr);
	//push_zstr (buf);
	
	// size of MULTI_EXPR
	sprintf (buf, "Size %d", ptr->Size);
	push_zstr (buf); 
	
	// handle to SYM_ENTRY
	sprintf (buf, "Hdl %d", se.hVal);
	push_zstr (buf);

	// flags of SYM_ENTRY
	switch (se.Flags) {
		case GRAPH_REF1:
			push_zstr("GRAPH_REF1");
			break;
		case GRAPH_REF2: 
			push_zstr("GRAPH_REF2");
			break;
		case SF_STAT_REF: 
			push_zstr("SF_STAT_REF");
			break;
		case SF_LOCK: 
			push_zstr("SF_LOCK");
			break;
		case SF_INUSE: 
			push_zstr("SF_INUSE");
			break;
		case SF_SELECTED:
			push_zstr("SF_SELECTED");
			break;
		case SF_RECEIVED: 
			push_zstr("SF_RECEIVED");
			break;
		case SF_FOLDER: 
			push_zstr("SF_FOLDER");
			break;
		case SF_INVIEW: 
			push_zstr("SF_INVIEW");
			break;
		case SF_EXTMEM: 
			push_zstr("SF_EXTMEM");
			break;
		case SF_EM_TWIN:
			push_zstr("SF_EM_TWIN");
			break;
		case SF_COLLAPSE: 
			push_zstr("SF_COLLAPSE");
			break;
		case SF_PARM:
			push_zstr("SF_PARM");
			break;
	}

	// version of SYM_ENTRY
	switch (se.Version) {
		case TV_TI_92:
			push_zstr ("TV_TI_92");
			break; 
		case TV_PARM:
			push_zstr ("TV_PARM");
			break; 		
		case TV_INTERNAL:
			push_zstr ("TV_INTERNAL");
			break; 		 
		case TV_SPAM:
			push_zstr ("TV_SPAM");
			break; 		 
		case TV_CRAM:
			push_zstr ("TV_CRAM");
			break; 		 
		case TV_3RDPARTYAPP:
			push_zstr ("TV_3RDPARTYAPP");
			break; 		
	}

	// name of SYM_ENTRY
	push_zstr ((char *)se.Name);
	
    push_quantum (LIST_TAG);
    
	cmdpause (top_estack);

    top_estack = old_top;
}

#endif
