// (c) Kevin Banks
// http://www.embedded.com/design/configurable-systems/4024443/The-Goertzel-Algorithm
// TI-89 frontend (c) Ralph Schenn

#include <tiams.h>
#include "Complex.h"
#include "Goertzel.h"
#include "Stack.h"

/* goertzel filter */
void goertzel(void)
{
    Access_AMS_Global_Variables;

    EStackIndex argptr = top_estack;
    volatile HANDLE hdl;
	double *ptr;
    unsigned short n;
    double fs, f;             

	// check number of parameters
	CheckNumArgs (3);
                        
	TRY	
		// read list of into array (in reverse order)
		// (automatically convert to float if necessary)
		n = GetFloatListArg (argptr, &hdl);
		
		argptr = next_expression_index (argptr);

		// get sampling frequency
		fs = estack_number_to_Float_inf (argptr);
		
		argptr = next_expression_index (argptr);
		
		// get searched frequency
		f = estack_number_to_Float_inf (argptr);
		
		// input list is not empty and 
		// length of average not longer than length of input list
		if (n > 0 && fs >= 0 && f >= 0 && f < fs / 2) {
			// dereference after the call to HeapAlloc 
			// and also after the call to HeapRealloc (called by GetFloatListArg)
			ptr = HeapDeref (hdl);
				
			// calculate signal power at searched frequency
			// push list to estack (in reverse order)
			push_complex_float (Goertzel(ptr, 
				n, 
				fs, 
				f));
		} else {
			ER_throw (ER_DOMAIN);		
		}
	FINALLY
		// HeapFreeIndir handles H_NULL correctly
		HeapFreeIndir ((HANDLE *)&hdl);
	ENDFINAL
}  

/* goertzel filter
input signal 
number of samples
sampling frequency (Hz)
searched frequency (Hz) 
http://www.embedded.com/design/configurable-systems/4024443/The-Goertzel-Algorithm */
Complex Goertzel(double x[], 
	unsigned short n, 
	double fs, 
	double f)
{
    Access_AMS_Global_Variables;  

	double w = (FLOATTAB[FPI_TWOPI] / n) * (0.5 + ((n * f) / fs));
	double cosine = cos (w);
	double sine = sin (w);
	double coeff = 2 * cosine;	
	double q0, q1 = 0., q2 = 0.;
	unsigned short i;
	Complex r;
	
	// loop over input samples
	for (i = 0; i < n; i++) {
		q0 = coeff * q1 - q2 + x[i];
		q2 = q1;
		q1 = q0;
	}
	
	r.real = q1 - q2 * cosine;
	r.imag = q2 * sine;
	return r;
	 
}

