/*
		Matrix functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Complex.h"

#ifndef _MATRIX_H_
#define _MATRIX_H_

/* convert two-dimensional index to one-dimensional index
number of columns
index of row
index of column */
#define IDX2D(cols, i, j) (i * cols + j)

/* initialize array of doubles */
void ArrInit(double a[], 
	short l);

/* initialize array of doubles */
void CArrInit(Complex a[], 
	short l);

/* partially initialize complex array */
void CInitMat(Complex a[],
	unsigned short start, 
	unsigned short end);

/* element-wise multiplication of complex arrays */
void CMatMatMul(Complex a[], 
	Complex b[], 
	short len);
	
/* divide complex array by complex scalar */
void CMatScalDiv(Complex a[],
	Complex b,
	short len);

/* pads array of complex from length l1 to length l2 */
unsigned short CPadPow2(HANDLE x,
	unsigned short oldLen,
	unsigned short newLen);

/* matrix product
matrix A
number of rows matrix A
number of colums matrix A
matrix B
number of colums matrix B
product of matices A B 
number of columns in A must be equal to the number of rows in B
speed comparison with builtin function depends on size of matrix and data type
even without being generally faster this function is usefull because it accepts 
the matrices stored as arrays instead of storage on estack 
https://en.wikipedia.org/wiki/Matrix_multiplication#Matrix_product_.28two_matrices.29 */
void MatProd(double *matA, 
	const unsigned short rowsA, 
	const unsigned short colsA, 
	double *matB, 
	const unsigned short colsB,
	double *prod);

#endif
