/*
		search functions
		
    Copyright (C) 2013  Ralph Schenn with support from Kevin Kofler

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Debug.h"
#include "MultiExpr.h"
#include "Stack.h"

/* check if array of handles to strings contains string 
return -1 if string was not found 
when string is found the index is returned */
/*short ArrayContStr(HANDLE strArray[], 
	short length, 
	const char *str)
{
	const char *strPtr = NULL;
	
	// array has already been allocated
	if (strArray != NULL) {
		// loop over all strings
		while (--length >= 0) {
			strPtr = HeapDeref (strArray[length]);
			
			// string was found in array
			if (strcmp (str, strPtr) == 0) {
				return length;
			}
		}
	}
	
	return -1;
}*/

#define NOT_FOUND 0

/* binary search
https://de.wikipedia.org/wiki/Bin%C3%A4re_Suche */
unsigned short BinSearch(EStackIndex list, 
	EStackIndex key)
{
	int length, comp, mid = 0, left = 0, right = 0;
	HANDLE elemHdl;
	EStackIndex *elemPtr;
	
	TRY
		// get array of pointers to list elements
		length = GetExpressionListArg (list, &elemHdl);
		
		elemPtr = HeapDeref (elemHdl);
		
		right = length - 1;
		
		// loop until interval to be searched is not empty
		while (left <= right) {
			// get half of list
			mid = left + ((right - left) / 2);
			
			comp = compare_expressions (elemPtr[mid], key);
			
			// element found
			if (comp == 0) {
				break;
			} else {
				if (comp > 0) {
					// search in left half
					right = mid - 1;
				} else {
					// search in right half
					left = mid + 1;
				}
			}
		}
	FINALLY
		HeapFreeIndir ((HANDLE *)&elemHdl);
	ENDFINAL
	
	// array index i 0 based while list index is 1 based
	return left <= right ? mid + 1 : NOT_FOUND;
}

/* search memory block subblock in block 
this works for arbitrary content without the need for \0 string termination 
case sensitive search 
http://stackoverflow.com/questions/9655565/search-block-of-memory-for-2-consecutive-chars */ 
const char *BlockContBlock(const char *memstart, // ok
	const char *string, 
	unsigned short memsize, 
	unsigned short stringsize)
{
	const char *p;
	
	for (p = memstart; p <= (memstart + memsize - stringsize); p++)
	{
	    // found
	    if (memcmp(p, string, stringsize) == 0) {
	    	return p;                        
	    }
	}
	
	// not found
	return NULL; 
}

/* count number of occurences of val in array 
array must not be null-terminated */
/*unsigned short CountChar(char array[], 
	unsigned short length, 
	char val) // ok
{
	unsigned short i, count = 0;
	
	// loop over characters in array
	for (i = 0; i < length; i++) {
		// match was found
		if (array[i] == val) {
			count++;
		}
	}
	
	return count;
}*/

#define BUF_LENGTH 30

/* recursive string search (commands for loop) */
/*void RecStrSe(SYM_ENTRY *symptr, 
	char *str, 
	char numArgs, 
	char *folder,
	unsigned short *numVars)
{
	// output buffer
	char out[BUF_LENGTH];	
	MULTI_EXPR *ptr;
	//char *debug;
	
	// display actual folder on status line
	//sprintf (out, "SEARCHING %s", varFolder);
	sprintf (out, "SEARCHING %s", folder);
	ST_helpMsg (out);
	
	// variable is not a folder
	if (IsFile (symptr)) { // && IsFuncOrProg (symptr)    		    
  		// variable contains search string
		ptr = HeapDeref (symptr->hVal);
		
		//debug = (char *)ptr->Expr; // ???
		
		if (BlockContBlock((const char *)ptr->Expr, 
			str, 
			ptr->Size, 
			strlen (str)) != NULL) {
			// add matiching variable to list
			sprintf (out, "%s\\%s", folder, symptr->Name);
			
			push_zstr (out);   	
		}
  	
  		(*numVars)++;
  	}
}*/

/* recursive variable search (commands for loop) */
void RecVarSe(SYM_ENTRY *symptr, 
	char *str, 
	char numArgs, 
	char *folder,
	unsigned short *numVars)
{
	char out[BUF_LENGTH];
	
	// object is file (not folder)
	if (IsFile (symptr)) {
		// name of file contains the search string    
		if (strstr((const char *)symptr->Name, str)) {
			// concat the path
			sprintf (out, "%s\\%s", 
				// SymFindFolderName is not working when searching 
				// only the variables in one folder
				numArgs == 1 ? (const char *)SymFindFoldername () : (const char *)folder, 
				symptr->Name);
				
			// add path to list
			push_zstr (out); 
		}
		
  		// increment number of searched variables
  		(*numVars)++;
  	}
}

/* recursive string or variable search */
void RecSearch(void (*funcPtr)(SYM_ENTRY *symptr, 
	char *str, 
	char numArgs, 
	char *folder, 
	unsigned short *numVars))
{
    Access_AMS_Global_Variables;

	// search folder name (ASCII format)
	char folderAscii[SYM_LEN + 1];

	// search string
	char str[BUF_LENGTH];	

	// flags for variable search
	unsigned short searchFlags;

	// pointer to variable in folder
	SYM_ENTRY *symptr;

	// number of variables which have been searched
	unsigned short numVars = 0;

	// output buffer
	char out[BUF_LENGTH];

	BYTE HomeFolder[] = {0, 127, 0};

	// argument pointer
	EStackIndex argptr = top_estack;
    
	// check number of arguments
	char numArgs = CheckNumArgsInt (1, 2);

	// search through all folders
	if (numArgs == 1) {
		// search all folders
		searchFlags = FO_RECURSE;
	// search only in  folder given in second argument
	} else {
		// get pointer to folder name from estack
		// copy the string to a variable independent from estack
		// (to avoid overwriting while pushing strings to the list)		
		GetStrnArgCopy (argptr, folderAscii, SYM_LEN + 1);
		
		// set pointer to search string
		argptr = next_expression_index (argptr);

		// search only in specified folder
		searchFlags = FO_NONE;
	}
	
	// get search string from estack
	// copy the string to a variable independent from estack
	// (to avoid overwriting while pushing strings to the list)		
	GetStrnArgCopy (argptr, str, BUF_LENGTH); 
			
	// lock home folder
	FolderOp (HomeFolder + 2, FL_LOCK);
	
	// if folder name was passed as first argument
	if (numArgs == 2) {
		// convert folder name in ASCII format to tokenized format
		TokenizeSymName ((BYTE *)folderAscii, TSF_PASS_ERRORS);
	} 
	
	// find first object in folder
	// if there have been passed two arguments the tokenized folder name is on top
	// if there has been passed only one argument the searched string is on top   
	// (but with FO_RECURSE the first parameter is ignored anyway)
	symptr = SymFindFirst (top_estack, searchFlags);  
	
	// if folder name was passed as first argument
	if (numArgs == 2) {
		// delete tokenized folder name from stack
		delete_expression (top_estack);
	} 

	// beginning of list
	push_quantum (END_TAG);

	// folder exists
	if (symptr) {
	    // loop over variables in folder(s)
	    do {
	    	// process the actual variable in folder
	    	(*funcPtr)(symptr, 
	    		str, 
	    		numArgs,
	    		folderAscii,
	    		&numVars);
	    		
	    // more variables exist
	    } while ((symptr = SymFindNext ()));
	// folder does not exist
	} else {
		ER_throwVar (ER_INVALID_FOLDER_NAME);        
	}
  	
	// unlock home folder
  	FolderOp (HomeFolder + 2, FL_UNLOCK);

	// end of list
	push_quantum (LIST_TAG);
  	
	// display number of variables searched on status line
	sprintf (out, "%d VARIABLES SEARCHED", numVars);
	ST_helpMsg (out);
	
	//#ifdef DEBUG
	//	PushStackAsMat ();
	//#endif             	
}
